%T2TL	conversion from Toeplitz to Toeplitz-like
%
%	[G,H]=t2tl(c,r,xi,eta) returns the generators of the Toeplitz matrix
%	having c as first column and r as first row, with respect to
%	Sylvester-type displacement 
%		Z_xi * T - T * Z_eta = G * H'

% 	Giuseppe Rodriguez,  21-01-07
% 	Email: rodriguez@unica.it

function [G,H] = t2tl(c,r,xi,eta)

if nargin < 3
	xi = 1;
	eta = -1;
end

if nargin<2,  r = c;  end

if abs(abs(xi)-1)>eps || abs(abs(eta)-1)>eps
	error( 'either xi or eta are not on the unit circle')
end

c = c(:);
r = r(:);
m = size(c,1);
n = size(r,1);
if c(1)~=r(1),  warning('Column wins diagonal conflict'),  end

t = [r(n:-1:2); c];
G = [[-eta*t(n);t(1:m-1)-eta*t(n+1:m+n-1)] [1;zeros(m-1,1)]];
H = [[zeros(n-1,1);1] conj([xi*t(m+n-1:-1:m+1)-t(n-1:-1:1);xi*t(m)])];
