%TL2CL	conversion from Toeplitz-like to Cauchy-like
%
%	[t,s,G,H]=tl2cl(xi,eta,Gt,Ht) returns the displacement diagonals and
%	the generators of the Cauchy-like matrix C corresponding to the
%	Toeplitz-like matrix having displacement Delta_{xi,eta} and generators
%	Gt and Ht, with respect to Sylvester-type displacement 
%		diag(t) * C - C * diag(s) = G * H'
%	[t,s,G,H,sumc,sumr]=tl2cl(xi,eta,Gt,Ht,c,r) returns also the sum of the
%	columns and the rows of C, given the first column and row of T 

% 	Giuseppe Rodriguez,  23-04-03
% 	Email: rodriguez@unica.it

function [t,s,G,H] = tl2cl(xi,eta,G,H,slow)

if nargin<5,  slow=0;  end

m = size(G,1);
n = size(H,1);

if slow
	Fxm = fmatn(m,xi);
	Dxm = dmat(m,xi,1);
	Fen = fmatn(n,eta);
	Den = dmat(n,eta,1);

	t = diag(Dxm);
	s = diag(Den);
	G = Fxm*G;
	H = Fen*H;
else
	t = dmat(m,xi);
	s = dmat(n,eta);
	G = fmatna(G,xi);
	H = fmatna(H,eta);
end

