/*
**	tplsmex.c
**
**	mex file per calcolare il prodotto matrice di Cauchy per vettore
**	mediante il complemento di Schur.
**	versione ottimizzata per la matrice mosaico B usata in tpls.m.
**	vedi tpls.m
**
**	versione senza variabili "complex", tutte le functions inline e
**	qualche altro trucco.
**
**	schurcmv() puo' ricevere piu' di un termine noto.
**
**	Giuseppe Rodriguez,  02-08-07
**	Email: rodriguez@unica.it
*/

#include <stdio.h>
#include <math.h>
#include "mex.h"

#define GR(i,j) gr[(i)+dim_g[0]*(j)]
#define GI(i,j) gi[(i)+dim_g[0]*(j)]
#define HR(i,j) hr[(i)+dim_h[0]*(j)]
#define HI(i,j) hi[(i)+dim_h[0]*(j)]

#if !defined(max)
#define	max(A, B)	((A) > (B) ? (A) : (B))
#endif

#if !defined(min)
#define	min(A, B)	((A) < (B) ? (A) : (B))
#endif


void schurc( int m, int n, double *tr, double *ti, double *sr, double *si,
	double *gr, double *gi, double *hr, double *hi,
	const int *dim_g, const int *dim_h, int piv, double *rankdef);
void schurcmv( int m, int n, double *tr, double *ti, double *sr, double *si,
	double *gr, double *gi, double *hr, double *hi,
	const int *dim_g, const int *dim_h, int kb);


void mexFunction(
	int		nlhs,
	mxArray		*plhs[],
	int		nrhs,
	const mxArray	*prhs[]
	)
{
	int job, m, n, piv, kb;
	double *tr, *ti, *sr, *si, *gr, *gi, *hr, *hi, *rankdef;
/*	int nod_t, nod_s, nod_g, nod_h; */
/*	const int *dim_t, *dim_s; */
	const int *dim_g, *dim_h;

	if (nrhs < 8) 
		mexErrMsgTxt( "Wrong number of input arguments");

	if (nlhs != 5) 
		mexErrMsgTxt( "Wrong number of output arguments");

	if ( !mxIsComplex(prhs[3]) || !mxIsComplex(prhs[4]) 
	   || !mxIsComplex(prhs[5]) || !mxIsComplex(prhs[6]) )
		mexErrMsgTxt( "Wrong type of input arguments");

	tr = mxGetPr( prhs[0]);
	job = lround(*tr);

	tr = mxGetPr( prhs[1]);
	sr = mxGetPr( prhs[2]);
	m = lround(*tr);
	n = lround(*sr);

	tr = mxGetPr( prhs[7]);
	piv = lround(*tr);
	if( job == 1 ) {
		tr = mxGetPr( prhs[8]);
		kb = lround(*tr);
	}

	plhs[0] = mxDuplicateArray(prhs[3]);
	plhs[1] = mxDuplicateArray(prhs[4]);
	plhs[2] = mxDuplicateArray(prhs[5]);
	plhs[3] = mxDuplicateArray(prhs[6]);
	plhs[4] = mxCreateNumericMatrix(1,1,mxDOUBLE_CLASS,mxREAL);

/*	nod_t = mxGetNumberOfDimensions(plhs[0]);
	nod_s = mxGetNumberOfDimensions(plhs[1]);
	nod_g = mxGetNumberOfDimensions(plhs[2]);
	nod_h = mxGetNumberOfDimensions(plhs[3]); 
	dim_t = mxGetDimensions(plhs[0]);
	dim_s = mxGetDimensions(plhs[1]); */
	dim_g = mxGetDimensions(plhs[2]);
	dim_h = mxGetDimensions(plhs[3]);

	tr = mxGetPr( plhs[0]);
	ti = mxGetPi( plhs[0]);
	sr = mxGetPr( plhs[1]);
	si = mxGetPi( plhs[1]);
	gr = mxGetPr( plhs[2]);
	gi = mxGetPi( plhs[2]);
	hr = mxGetPr( plhs[3]);
	hi = mxGetPi( plhs[3]);
	rankdef = mxGetPr( plhs[4]);

	switch( job )
	{
	case 0:
		schurc(m,n,tr,ti,sr,si,gr,gi,hr,hi,dim_g,dim_h,piv,rankdef);
		break;
	case 1:
		schurcmv(m,n,tr,ti,sr,si,gr,gi,hr,hi,dim_g,dim_h,kb);
		*rankdef = 0.0;
		break;
	default:
		mexErrMsgTxt( "First arguments is wrong");
	}
}


void schurc( int m, int n, double *tr, double *ti, double *sr, double *si,
	double *gr, double *gi, double *hr, double *hi,
	const int *dim_g, const int *dim_h, int piv, double *rankdef)
{
	int n1, n2, *p, *i23, imax, k, i, j, k1;
	double *lr, *li, *ur, *ui, *d22r, *d22i, *m23r, *m23i;
	double zr, zi, wr, wi, rho, rholk, maxlk, modlk, eps;

	*rankdef = 0.0;
	eps = mxGetEps();

	n1 = m+2*n;
	n2 = 2*m+n;

	p = mxCalloc( m+n, sizeof(int));
	lr = mxCalloc( n1, sizeof(double));
	li = mxCalloc( n1, sizeof(double));
	ur = mxCalloc( n2, sizeof(double));
	ui = mxCalloc( n2, sizeof(double));
	d22r = mxCalloc( n, sizeof(double));
	d22i = mxCalloc( n, sizeof(double));
	m23r = mxCalloc( n, sizeof(double));
	m23i = mxCalloc( n, sizeof(double));
	i23 = mxCalloc( n, sizeof(int));

	for( i=0; i<n; i++) {
		d22r[i] = 0.;  d22i[i] = 0.;
		zr = 0.;  zi = 0.;
		for( j=0; j<dim_g[1]; j++) {
			zr = zr + (GR(m+i,j)*HR(m+n,j) + GI(m+i,j)*HI(m+n,j));
			zi = zi + (-GR(m+i,j)*HI(m+n,j) + GI(m+i,j)*HR(m+n,j));
		}
		wr = tr[m+i] - sr[m+n];  wi = ti[m+i] - si[m+n];
		if ( (wr == 0.) && (wi == 0.) )
			mexErrMsgTxt( "Null denominator");
		rho = 1./(wr*wr+wi*wi);
		m23r[i] = (zr*wr+zi*wi)*rho;
		m23i[i] = (zi*wr-zr*wi)*rho;
		i23[i] = 0;
		p[i] = i;
	}
	for( i=n; i<m+n; i++)
		p[i] = i;

	for( k=0; k<m; k++) {
		lr[k] = 1.;  li[k] = 0.;
		rholk = 1.;  imax = k;
		for( i=m; i<m+n; i++) {
			zr = 0.;  zi = 0.;
			for( j=0; j<dim_g[1]; j++) {
				zr = zr + (GR(i,j)*HR(k,j) + GI(i,j)*HI(k,j));
				zi = zi + (-GR(i,j)*HI(k,j) + GI(i,j)*HR(k,j));
			}
			wr = tr[i] - sr[k];  wi = ti[i] - si[k];
			if ( (wr == 0.) && (wi == 0.) )
				mexErrMsgTxt( "Null denominator");
			rho = 1./(wr*wr+wi*wi);
			lr[i] = (zr*wr+zi*wi)*rho;
			li[i] = (zi*wr-zr*wi)*rho;
			/* zr = lr[i]*lr[i] + li[i]*li[i]; */
			zr = fabs(lr[i]) + fabs(li[i]);
			if( zr > rholk) {
				rholk = zr;  imax = i;
			}
		}
		if( piv && (imax > k) ) {
			if( p[imax] == imax ) {
				k1 = imax;
				ur[imax] = d22r[imax-m];
				ui[imax] = d22i[imax-m];
			} else {
				k1 = m + n + i23[imax-m];
				ur[k1] = m23r[imax-m];
				ui[k1] = m23i[imax-m];
			}
			for( i=k; i<k1; i++) {
				zr = 0.;  zi = 0.;
				for( j=0; j<dim_g[1]; j++) {
					zr = zr + (GR(imax,j)*HR(i,j) + 
							GI(imax,j)*HI(i,j));
					zi = zi + (-GR(imax,j)*HI(i,j) + 
							GI(imax,j)*HR(i,j));
				}
				wr = tr[imax] - sr[i];  wi = ti[imax] - si[i];
				if ( (wr == 0.) && (wi == 0.) )
					mexErrMsgTxt( "Null denominator");
				rho = 1./(wr*wr+wi*wi);
				ur[i] = (zr*wr+zi*wi)*rho;
				ui[i] = (zi*wr-zr*wi)*rho;
			}
			for( i=k1+1; i<n2; i++) {
				zr = 0.;  zi = 0.;
				for( j=0; j<dim_g[1]; j++) {
					zr = zr + (GR(imax,j)*HR(i,j) + 
							GI(imax,j)*HI(i,j));
					zi = zi + (-GR(imax,j)*HI(i,j) + 
							GI(imax,j)*HR(i,j));
				}
				wr = tr[imax] - sr[i];  wi = ti[imax] - si[i];
				if ( (wr == 0.) && (wi == 0.) )
					mexErrMsgTxt( "Null denominator");
				rho = 1./(wr*wr+wi*wi);
				ur[i] = (zr*wr+zi*wi)*rho;
				ui[i] = (zi*wr-zr*wi)*rho;
			}
			zr = 0.;  zi = 0.;
			for( j=0; j<dim_g[1]; j++) {
				zr = zr + (GR(k,j)*HR(imax,j) +
						GI(k,j)*HI(imax,j));
				zi = zi + (-GR(k,j)*HI(imax,j) +
						GI(k,j)*HR(imax,j));
			}
			wr = tr[k] - sr[imax];  wi = ti[k] - si[imax];
			if ( (wr == 0.) && (wi == 0.) )
				mexErrMsgTxt( "Null denominator");
			rho = 1./(wr*wr+wi*wi);
			d22r[imax-m] = (zr*wr+zi*wi)*rho;
			d22i[imax-m] = (zi*wr-zr*wi)*rho;
			m23r[imax-m] = 0.;  m23i[imax-m] = 0.;
			i23[imax-m] = k;
			zr = lr[k];  lr[k] = lr[imax];  lr[imax] = zr;
			zi = li[k];  li[k] = li[imax];  li[imax] = zi;
			zr = tr[k];  tr[k] = tr[imax];  tr[imax] = zr;
			zi = ti[k];  ti[k] = ti[imax];  ti[imax] = zi;
			for( j=0; j<dim_g[1]; j++) {
				zr = GR(k,j);  GR(k,j) = GR(imax,j);  
					GR(imax,j) = zr;
				zi = GI(k,j);  GI(k,j) = GI(imax,j);  
					GI(imax,j) = zi;
			}
			k1 = p[k];  p[k] = p[imax];  p[imax] = k1;
		} else {
			ur[k] = 1.;  ui[k] = 0.;
			for( i=k+1; i<m; i++) {
				ur[i] = 0.;  ui[i] = 0.;
			}
			for( i=m; i<m+n; i++) {
				zr = 0.;  zi = 0.;
				for( j=0; j<dim_g[1]; j++) {
					zr = zr + (GR(k,j)*HR(i,j) +
							GI(k,j)*HI(i,j));
					zi = zi + (-GR(k,j)*HI(i,j) +
							GI(k,j)*HR(i,j));
				}
				wr = tr[k] - sr[i];  wi = ti[k] - si[i];
				if ( (wr == 0.) && (wi == 0.) )
					mexErrMsgTxt( "Null denominator");
				rho = 1./(wr*wr+wi*wi);
				ur[i] = (zr*wr+zi*wi)*rho;
				ui[i] = (zi*wr-zr*wi)*rho;
			}
			for( i=m+n; i<n2; i++) {
				ur[i] = 0.;  ui[i] = 0.;
			}
		}
		rholk = 1./(lr[k]*lr[k]+li[k]*li[k]);
		for( i=m; i<m+n; i++) {
			zr = (lr[i]*lr[k]+li[i]*li[k])*rholk;
			zi = (li[i]*lr[k]-lr[i]*li[k])*rholk;
			lr[i] = zr;  li[i] = zi;
			d22r[i-m] = d22r[i-m] - (lr[i]*ur[i]-li[i]*ui[i]);
			d22i[i-m] = d22i[i-m] - (lr[i]*ui[i]+li[i]*ur[i]);
			m23r[i-m] = m23r[i-m] - (lr[i]*ur[m+n+i23[i-m]] - 
						li[i]*ui[m+n+i23[i-m]]);
			m23i[i-m] = m23i[i-m] - (lr[i]*ui[m+n+i23[i-m]] + 
						li[i]*ur[m+n+i23[i-m]]);
			for( j=0; j<dim_g[1]; j++) {
				GR(i,j) = GR(i,j) - (lr[i]*GR(k,j) -
							li[i]*GI(k,j));
				GI(i,j) = GI(i,j) - (lr[i]*GI(k,j) +
							li[i]*GR(k,j));
			}
		}
		for( i=k+1; i<n2; i++) {
			zr = (ur[i]*lr[k]+ui[i]*li[k])*rholk;
			zi = (ui[i]*lr[k]-ur[i]*li[k])*rholk;
			ur[i] = zr;  ui[i] = zi;
			for( j=0; j<dim_h[1]; j++) {
				HR(i,j) = HR(i,j) - (ur[i]*HR(k,j) +
							ui[i]*HI(k,j));
				HI(i,j) = HI(i,j) - (ur[i]*HI(k,j) -
							ui[i]*HR(k,j));
			}
		}
	}

	maxlk = 0.0;
	for( k=m; k<m+n; k++) {
		lr[k] = d22r[k-m];  li[k] = d22i[k-m];
		for( i=k+1; i<n+k; i++) {
			zr = 0.;  zi = 0.;
			for( j=0; j<dim_g[1]; j++) {
				zr = zr + (GR(i,j)*HR(k,j) + GI(i,j)*HI(k,j));
				zi = zi + (-GR(i,j)*HI(k,j) + GI(i,j)*HR(k,j));
			}
			wr = tr[i] - sr[k];  wi = ti[i] - si[k];
			if ( (wr == 0.) && (wi == 0.) )
				mexErrMsgTxt( "Null denominator");
			rho = 1./(wr*wr+wi*wi);
			lr[i] = (zr*wr+zi*wi)*rho;
			li[i] = (zi*wr-zr*wi)*rho;
		}
		lr[n+k] = 1.;  li[n+k] = 0.;
		/* rholk = lr[k]*lr[k] + li[k]*li[k];  imax = k; */
		rholk = fabs(lr[k]) + fabs(li[k]);  imax = k;
		for( i=k+1; i<m+n; i++) {
			/* zr = lr[i]*lr[i] + li[i]*li[i]; */
			zr = fabs(lr[i]) + fabs(li[i]);
			if( zr > rholk) {
				rholk = zr;  imax = i;
			}
		}
		if( piv && (imax > k) ) {
			if( p[imax] == imax ) {
				k1 = imax;
				ur[k1] = d22r[imax-m];
				ui[k1] = d22i[imax-m];
			} else {
				k1 = m + n + i23[imax-m];
				ur[k1] = m23r[imax-m];
				ui[k1] = m23i[imax-m];
			}
			for( i=k; i<k1; i++) {
				zr = 0.;  zi = 0.;
				for( j=0; j<dim_g[1]; j++) {
					zr = zr + (GR(imax,j)*HR(i,j) +
							GI(imax,j)*HI(i,j));
					zi = zi + (-GR(imax,j)*HI(i,j) +
							GI(imax,j)*HR(i,j));
				}
				wr = tr[imax] - sr[i];  wi = ti[imax] - si[i];
				if ( (wr == 0.) && (wi == 0.) )
					mexErrMsgTxt( "Null denominator");
				rho = 1./(wr*wr+wi*wi);
				ur[i] = (zr*wr+zi*wi)*rho;
				ui[i] = (zi*wr-zr*wi)*rho;
			}
			for( i=k1+1; i<n2; i++) {
				zr = 0.;  zi = 0.;
				for( j=0; j<dim_g[1]; j++) {
					zr = zr + (GR(imax,j)*HR(i,j) +
							GI(imax,j)*HI(i,j));
					zi = zi + (-GR(imax,j)*HI(i,j) +
							GI(imax,j)*HR(i,j));
				}
				wr = tr[imax] - sr[i];  wi = ti[imax] - si[i];
				if ( (wr == 0.) && (wi == 0.) )
					mexErrMsgTxt( "Null denominator");
				rho = 1./(wr*wr+wi*wi);
				ur[i] = (zr*wr+zi*wi)*rho;
				ui[i] = (zi*wr-zr*wi)*rho;
			}
			zr = 0.;  zi = 0.;
			for( j=0; j<dim_g[1]; j++) {
				zr = zr + (GR(k,j)*HR(imax,j) +
						GI(k,j)*HI(imax,j));
				zi = zi + (-GR(k,j)*HI(imax,j) +
						GI(k,j)*HR(imax,j));
			}
			wr = tr[k] - sr[imax];  wi = ti[k] - si[imax];
			if ( (wr == 0.) && (wi == 0.) )
				mexErrMsgTxt( "Null denominator");
			rho = 1./(wr*wr+wi*wi);
			d22r[imax-m] = (zr*wr+zi*wi)*rho;
			d22i[imax-m] = (zi*wr-zr*wi)*rho;
			m23r[imax-m] = m23r[k-m];  m23i[imax-m] = m23i[k-m];
			i23[imax-m] = i23[k-m] ;
			zr = lr[k];  lr[k] = lr[imax];  lr[imax] = zr;
			zi = li[k];  li[k] = li[imax];  li[imax] = zi;
			zr = tr[k];  tr[k] = tr[imax];  tr[imax] = zr;
			zi = ti[k];  ti[k] = ti[imax];  ti[imax] = zi;
			for( j=0; j<dim_g[1]; j++) {
				zr = GR(k,j);  GR(k,j) = GR(imax,j);  
					GR(imax,j) = zr;
				zi = GI(k,j);  GI(k,j) = GI(imax,j);  
					GI(imax,j) = zi;
			}
			k1 = p[k];  p[k] = p[imax];  p[imax] = k1;
		} else {
			if( p[k] == k ) {
				k1 = k;
				ur[k1] = d22r[k-m];  ui[k1] = d22i[k-m];
			} else {
				k1 = m + n + i23[k-m];
				ur[k1] = m23r[k-m];  ui[k1] = m23i[k-m];
			}
			if( k1 > k )
				for( i=k; i<k1; i++) {
					zr = 0.;  zi = 0.;
					for( j=0; j<dim_g[1]; j++) {
						zr = zr + (GR(k,j)*HR(i,j) +
							GI(k,j)*HI(i,j));
						zi = zi + (-GR(k,j)*HI(i,j) +
							GI(k,j)*HR(i,j));
					}
					wr = tr[k] - sr[i];  wi = ti[k] - si[i];
					if ( (wr == 0.) && (wi == 0.) )
						mexErrMsgTxt( "Null denominator");
					rho = 1./(wr*wr+wi*wi);
					ur[i] = (zr*wr+zi*wi)*rho;
					ui[i] = (zi*wr-zr*wi)*rho;
				}
			for( i=k1+1; i<n2; i++) {
				zr = 0.;  zi = 0.;
				for( j=0; j<dim_g[1]; j++) {
					zr = zr + (GR(k,j)*HR(i,j) +
							GI(k,j)*HI(i,j));
					zi = zi + (-GR(k,j)*HI(i,j) +
							GI(k,j)*HR(i,j));
				}
				wr = tr[k] - sr[i];  wi = ti[k] - si[i];
				if ( (wr == 0.) && (wi == 0.) )
					mexErrMsgTxt( "Null denominator");
				rho = 1./(wr*wr+wi*wi);
				ur[i] = (zr*wr+zi*wi)*rho;
				ui[i] = (zi*wr-zr*wi)*rho;
			}
		}
		rholk = 1./(lr[k]*lr[k]+li[k]*li[k]);
		modlk = fabs(lr[k]) + fabs(li[k]);
		if( modlk > maxlk )
			maxlk = modlk;
		if( modlk < eps*maxlk ) {
			*rankdef = 1.0;
			return;
		}
		for( i=k+1; i<n+k+1; i++) {
			zr = (lr[i]*lr[k]+li[i]*li[k])*rholk;
			zi = (li[i]*lr[k]-lr[i]*li[k])*rholk;
			lr[i] = zr;  li[i] = zi;
			for( j=0; j<dim_g[1]; j++) {
				GR(i,j) = GR(i,j) - (lr[i]*GR(k,j) -
							li[i]*GI(k,j));
				GI(i,j) = GI(i,j) - (lr[i]*GI(k,j) +
							li[i]*GR(k,j));
			}
		}
		for( i=k+1; i<m+n; i++) {
			d22r[i-m] = d22r[i-m] - (lr[i]*ur[i]-li[i]*ui[i]);
			d22i[i-m] = d22i[i-m] - (lr[i]*ui[i]+li[i]*ur[i]);
			m23r[i-m] = m23r[i-m] - (lr[i]*ur[m+n+i23[i-m]] - 
						li[i]*ui[m+n+i23[i-m]]);
			m23i[i-m] = m23i[i-m] - (lr[i]*ui[m+n+i23[i-m]] + 
						li[i]*ur[m+n+i23[i-m]]);
		}
		for( i=k+1; i<n2; i++) {
			zr = (ur[i]*lr[k]+ui[i]*li[k])*rholk;
			zi = (ui[i]*lr[k]-ur[i]*li[k])*rholk;
			ur[i] = zr;  ui[i] = zi;
			for( j=0; j<dim_h[1]; j++) {
				HR(i,j) = HR(i,j) - (ur[i]*HR(k,j) +
							ui[i]*HI(k,j));
				HI(i,j) = HI(i,j) - (ur[i]*HI(k,j) -
							ui[i]*HR(k,j));
			}
			
		}
	}

	mxFree(p);
	mxFree(lr);
	mxFree(li);
	mxFree(ur);
	mxFree(ui);
	mxFree(d22r);
	mxFree(d22i);
	mxFree(m23r);
	mxFree(m23i);
	mxFree(i23);
}


void schurcmv( int m, int n, double *tr, double *ti, double *sr, double *si,
	double *gr, double *gi, double *hr, double *hi,
	const int *dim_g, const int *dim_h, int kb)
{
	int k, i, j;
	double *lr, *li, *ur, *ui, zr, zi, wr, wi, rho;

	lr = mxCalloc( m+n, sizeof(double));
	li = mxCalloc( m+n, sizeof(double));
	ur = mxCalloc( m+kb, sizeof(double));
	ui = mxCalloc( m+kb, sizeof(double));

	for( k=0; k<m; k++) {
		for( i=m; i<m+n; i++) {
			zr = 0.;  zi = 0.;
			for( j=0; j<dim_g[1]; j++) {
				zr = zr + (GR(i,j)*HR(k,j) + GI(i,j)*HI(k,j));
				zi = zi + (-GR(i,j)*HI(k,j) + GI(i,j)*HR(k,j));
			}
			wr = tr[i] - sr[k];  wi = ti[i] - si[k];
			if ( (wr == 0.) && (wi == 0.) )
				mexErrMsgTxt( "Null denominator");
			rho = 1./(wr*wr+wi*wi);
			lr[i] = (zr*wr+zi*wi)*rho;
			li[i] = (zi*wr-zr*wi)*rho;
		}
		for( i=m; i<m+kb; i++) {
			zr = 0.;  zi = 0.;
			for( j=0; j<dim_g[1]; j++) {
				zr = zr + (GR(k,j)*HR(i,j) + GI(k,j)*HI(i,j));
				zi = zi + (-GR(k,j)*HI(i,j) + GI(k,j)*HR(i,j));
			}
			wr = tr[k] - sr[i];  wi = ti[k] - si[i];
			if ( (wr == 0.) && (wi == 0.) )
				mexErrMsgTxt( "Null denominator");
			rho = 1./(wr*wr+wi*wi);
			ur[i] = (zr*wr+zi*wi)*rho;
			ui[i] = (zi*wr-zr*wi)*rho;
		}
		for( j=0; j<dim_g[1]; j++) {
			for( i=m; i<m+n; i++) {
				GR(i,j) = GR(i,j) + (lr[i]*GR(k,j) -
							li[i]*GI(k,j));
				GI(i,j) = GI(i,j) + (lr[i]*GI(k,j) +
							li[i]*GR(k,j));
			}
			for( i=m; i<m+kb; i++) {
				HR(i,j) = HR(i,j) + (ur[i]*HR(k,j) + 
							ui[i]*HI(k,j));
				HI(i,j) = HI(i,j) + (ur[i]*HI(k,j) - 
							ui[i]*HR(k,j));
			}
		}
	}

	mxFree(lr);
	mxFree(li);
	mxFree(ur);
	mxFree(ui);
}

